import random
import uuid

import pandas as pd
import pytest
import torch
import torchvision.models as models
from boto3 import session
from torch.utils.data import DataLoader

import daft
from daft.catalog.__glue import GlueCatalog


@pytest.mark.integration()
def test_dataloading_from_glue_iceberg(pytestconfig):
    if pytestconfig.getoption("--credentials") is not True:
        pytest.skip("Test can only run in a credentialled environment, and when run with the `--credentials` flag")

    sess = session.Session()
    catalog = GlueCatalog.from_session(name="glue_catalog", session=sess)

    images_table = catalog.get_table("glue_iceberg_test.coco_images")
    images_df = images_table.read()  # noqa: F841

    categories_table = catalog.get_table("glue_iceberg_test.coco_categories")
    categories_df = categories_table.read().with_column_renamed("id", "category_id")  # noqa: F841

    annotations_table = catalog.get_table("glue_iceberg_test.coco_annotations")
    annotations_df = annotations_table.read().with_column_renamed("id", "annotation_id")  # noqa: F841

    df = daft.sql("""
        SELECT
            id, annotation_id, bbox, bucket_url
        FROM
            images_df
        JOIN
            annotations_df ON id = image_id
        JOIN
            categories_df USING (category_id)
        WHERE
            supercategory = 'person'
    """)
    df = df.with_column("image", df["bucket_url"].url.download().image.decode())
    df = df.with_column("cropped", df["image"].image.crop(df["bbox"]))
    df = df.with_column("thumbnail", df["cropped"].image.resize(64, 64))
    df = df.with_column("image_tensor", df["thumbnail"]).select("id", "annotation_id", "image_tensor").limit(100)

    df.collect()
    assert df.count_rows() == 100

    model = models.resnet18()
    model = torch.nn.Sequential(*list(model.children())[:-1])
    model.eval()
    device = torch.device("cpu")
    model.to(device)

    torch_dataset = df.to_torch_iter_dataset()
    dataloader = DataLoader(torch_dataset, batch_size=16)

    embeddings = []
    ids = []
    annotation_ids = []

    for batch in iter(dataloader):
        image_tensor = batch["image_tensor"].to(device)
        image_tensor = image_tensor.permute(0, 3, 1, 2).float() / 255.0
        outputs = model(image_tensor)
        outputs = outputs.view(outputs.size(0), -1)
        embeddings.append(outputs.cpu())
        ids.extend(batch["id"])
        annotation_ids.extend(batch["annotation_id"])

    all_embeddings = torch.cat(embeddings, dim=0)

    assert all_embeddings.shape[0] == len(ids), "Number of embeddings should match number of ids"
    assert all_embeddings.shape[1] == 512, "ResNet18 should produce 512-dimensional embeddings"


@pytest.mark.integration()
def test_pandas_to_daft_to_glue_iceberg(pytestconfig):
    if pytestconfig.getoption("--credentials") is not True:
        pytest.skip("Test requires AWS credentials and `--credentials` flag")

    # Create random data in pandas.
    num_rows = 10
    df_pd = pd.DataFrame(
        {
            "id": [str(uuid.uuid4()) for _ in range(num_rows)],
            "double_col": [random.random() for _ in range(num_rows)],
            "int_col": [random.randint(0, 3) for _ in range(num_rows)],
        }
    )

    # Convert to daft DataFrame.
    df = daft.from_pandas(df_pd)

    # Overwrite data in Glue Iceberg table.
    sess = session.Session()
    catalog = GlueCatalog.from_session(name="glue_catalog", session=sess)
    table = catalog.get_table("glue_iceberg_test.pandas_to_glue_table")
    table.write(df, mode="overwrite")

    # Validate that we wrote the data correctly.
    read_back = table.read()
    read_back_pd = read_back.sort("id").to_pandas().reset_index(drop=True)
    df_pd = df_pd.sort_values("id").reset_index(drop=True)
    assert read_back_pd.equals(df_pd), "Did not write data correctly to Glue Iceberg table"



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic, Iterator, TypeVar

from daft.recordbatch import MicroPartition
from daft.schema import Schema

T = TypeVar("T")


@dataclass
class WriteOutput(Generic[T]):
    """Wrapper for output of the DataSink's `.write()` method.

    Attributes:
        output: The actual output from the write operation
        bytes_written: Size of the written data in bytes
        rows_written: Number of rows written
    """

    output: T
    bytes_written: int
    rows_written: int


class DataSink(ABC, Generic[T]):
    """Interface for writing data to a sink that is not built-in.

    When a DataFrame is written using the `.write_sink()` method, the following sequence occurs:

    1. The sink's `.start()` method is called once at the beginning of the write process.
    2. The DataFrame is executed, and its output is split into micropartitions.
    3. The sink's `.write()` method is invoked on each micropartition, potentially in parallel
       and distributed across multiple tasks or workers.
    4. After all writes complete, the resulting `WriteOutput` objects are gathered on a single node.
    5. The `.finalize()` method is then called with all write outputs to produce a final `MicroPartition`.

    Warning:
        This API is early in its development and is subject to change.
    """

    def name(self) -> str:
        """Optional custom sink name."""
        return "User-defined Data Sink"

    @abstractmethod
    def schema(self) -> Schema:
        """The expected schema for the micropartition returned by the `.finalize()` method of this DataSink.

        If this given schema does not match the actual schema of the micropartition at runtime, we throw an error.
        """
        raise NotImplementedError

    def start(self) -> None:
        """Optional callback for when a write operation begins.

        For example, this can be used to initialize resources, open connections, start a transaction etc.
        The default implementation does nothing.
        """
        pass

    @abstractmethod
    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteOutput[T]]:
        """Writes a stream of micropartitions to the sink.

        This method should handle the ingestion of each micropartition and yield a result
        (e.g. metadata) for each successful write.

        Args:
            micropartitions (Iterator[MicroPartition]): An iterator of micropartitions to be written.

        Returns:
            Iterator[WriteOutput[T]]: An iterator of write results wrapped in a WriteOutput.
        """
        raise NotImplementedError

    @abstractmethod
    def finalize(self, write_outputs: list[WriteOutput[T]]) -> MicroPartition:
        """Finalizes the write process and returns a resulting micropartition.

        For example, this can be used to merge, summarize, or commit the results of individual writes
        into a single output micropartition.

        Args:
            write_outputs (list[WriteOutput[T]]): The list of results from the calls to `.write()`.

        Returns:
            MicroPartition: A final, single micropartition representing the result of all writes.
        """
        raise NotImplementedError



"""An in-memory implementation for the daft catalog abstractions."""

from __future__ import annotations

from typing import Any

from daft.catalog import Catalog, Identifier, NotFoundError, Properties, Schema, Table
from daft.dataframe.dataframe import DataFrame


class MemoryCatalog(Catalog):
    """An in-memory catalog is backed by a dictionary."""

    _name: str
    _tables: dict[str, Table]

    def __init__(self, name: str, tables: list[Table] = []):
        self._name = name
        self._tables = {t.name: t for t in tables}

    @property
    def name(self) -> str:
        return self._name

    @staticmethod
    def _from_pydict(name: str, tables: dict[str, object]) -> MemoryCatalog:
        mem = MemoryCatalog(name)
        for path, source in tables.items():
            ident = Identifier.from_str(path)
            table = MemoryTable._from_pyobject(str(ident[-1]), source)
            mem._tables[path] = table
        return mem

    ###
    # create_*
    ###

    def create_namespace(self, identifier: Identifier | str) -> None:
        raise NotImplementedError("Memory create_namespace not yet supported.")

    def create_table(
        self,
        identifier: Identifier | str,
        source: Schema | DataFrame,
        properties: Properties | None = None,
    ) -> Table:
        raise NotImplementedError("Memory create_table not yet supported.")

    ###
    # drop_*
    ###

    def drop_namespace(self, identifier: Identifier | str) -> None:
        raise NotImplementedError("Memory drop_namespace not yet supported.")

    def drop_table(self, identifier: Identifier | str) -> None:
        raise NotImplementedError("Memory drop_table not yet supported.")

    ###
    # has_*
    ###

    def has_namespace(self, identifier: Identifier | str) -> bool:
        prefix = str(identifier)
        for ident in self._tables.keys():
            if ident.startswith(prefix):
                return True
        return False

    def has_table(self, identifier: Identifier | str) -> bool:
        return str(identifier) in self._tables

    ###
    # list_*
    ###

    def list_namespaces(self, pattern: str | None = None) -> list[Identifier]:
        namespaces = set()
        for path in self._tables.keys():
            if pattern is not None and not path.startswith(pattern):
                continue  # did not match pattern
            split = path.rfind(".")
            if split == -1:
                continue  # does not have a namespace
            namespaces.add(path[:split])
        return [Identifier.from_str(ns) for ns in namespaces]

    def list_tables(self, pattern: str | None = None) -> list[str]:
        if pattern is None:
            return list(self._tables.keys())
        return [path for path in self._tables.keys() if path.startswith(pattern)]

    ###
    # get_*
    ###

    def get_table(self, identifier: str | Identifier) -> Table:
        path = str(identifier)
        if path not in self._tables:
            raise NotFoundError(f"Table {path} does not exist.")
        return self._tables[path]


class MemoryTable(Table):
    """An in-memory table holds a reference to an existing dataframe."""

    _name: str
    _inner: DataFrame

    def __init__(self, name: str, inner: DataFrame):
        self._name = name
        self._inner = inner

    @property
    def name(self) -> str:
        return self._name

    @staticmethod
    def _from_pyobject(name: str, source: object) -> Table:
        """Returns a Daft Table from a supported object type or raises an error."""
        if isinstance(source, Table):
            # we want to rename and create an immutable view from this external table
            return Table.from_df(name, source.read())
        elif isinstance(source, DataFrame):
            return Table.from_df(name, source)
        elif isinstance(source, dict):
            return Table.from_df(name, DataFrame._from_pydict(source))
        else:
            raise ValueError(f"Unsupported table source {type(source)}")

    ###
    # read methods
    ###

    def read(self, **options: Any) -> DataFrame:
        return self._inner

    ###
    # write methods
    ###

    def write(self, df: DataFrame, mode: str = "append", **options: Any) -> None:
        raise NotImplementedError("Writes to in-memory tables are not yet supported.")
